package pw.stadnik.searcher.agent;

import java.sql.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import pw.stadnik.searcher.DataAccess;

/**
 * 
 * This class provides access to database
 * for agents.
 * 
 * @author Jakub Stadnik
 *
 */
public class AgentDataAccess extends DataAccess
{

	/**
	 * Returns search engines associated with given
	 * query and algorithm. Associated means that engines 
	 * that are to be returned are those which are in
	 * the knowledge base of a given agent. If the knowledge
	 * base does not contain entry for particular configuration
	 * (query, algorithm) search engines are
	 * returned in unbiased way, i.e. the list is shuffled
	 * before being returned. Method does not return
	 * the actual search engine objects but rather their
	 * identifiers.
	 * 
	 * @param query query for which to return the search engines
	 * @param algorithm algorithm for which to return the search engines
	 * @return list of search engines
	 */
	public List<Integer> getSearchEngines(String query, String algorithm)
	{
		List<Integer> engineIds = new ArrayList<Integer>();
		
		try
		{
			String sql = "SELECT search_engine_id FROM query_data " +
					"WHERE query = ? " +
					"AND algorithm_id IN (SELECT id FROM algorithm WHERE name = ?) " +
					"ORDER BY weight";
			
			PreparedStatement statement = connection.prepareStatement(sql);
			
			statement.setString(1, query);
			statement.setString(2, algorithm);
			
			ResultSet resultSet = statement.executeQuery();
			
			while(resultSet.next())
			{
				engineIds.add(resultSet.getInt(1));
			}
			
			statement.close();
			resultSet.close();
			if(engineIds.size() == 0)
			{
				sql = "SELECT id FROM search_engine WHERE is_enabled = 1";
				
				statement = connection.prepareStatement(sql);
				
				resultSet = statement.executeQuery();
				
				while(resultSet.next())
				{
					engineIds.add(resultSet.getInt(1));
				}
				
				statement.close();
				resultSet.close();
				Collections.shuffle(engineIds);
			}	
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return engineIds;
	}
	
	/**
	 * Updates knowledge base with particular configuration
	 * (query, search engine, weight, algorithm).
	 * 
	 * @param query query for which the knowledge base should be updated
	 * @param searchEngineId search engine identifier for which the knowledge base should be updated
	 * @param weight weight with which the knowledge base should be updated
	 * @param algorithm algorithm for which the knowledge base should be updated
	 */
	public void updateKnowledgeBase(String query, Integer searchEngineId, Float weight, String algorithm)
	{
		try
		{
			
			String sql = "SELECT id FROM query_data WHERE search_engine_id = ? " +
					"AND query = ?" +
					" AND algorithm_id = (SELECT id FROM algorithm WHERE name = ?)";
			
			PreparedStatement statement = connection.prepareStatement(sql);
			
			statement.setInt(1, searchEngineId);
			statement.setString(2, query);
			statement.setString(3, algorithm);
			
			ResultSet resultSet = statement.executeQuery();
			
			if(resultSet.next())
			{
				sql = "UPDATE query_data SET weight = ((weight * query_count) + ?) / (query_count + 1), " +
						"query_count = query_count + 1 " +
						"WHERE search_engine_id = ? AND query = ? " +
						"AND algorithm_id IN (SELECT id FROM algorithm WHERE name = ?)";
			}
			else
			{
				sql = "INSERT INTO query_data(weight, search_engine_id, query, algorithm_id, query_count) " +
						"SELECT ?, ?, ?, id, 1 FROM algorithm WHERE name = ?";
			}
			
			PreparedStatement ps = connection.prepareStatement(sql);
			ps.setFloat(1, weight);
			ps.setInt(2, searchEngineId);
			ps.setString(3, query);
			ps.setString(4, algorithm);
			
			ps.executeUpdate();
			
			ps.close();
			
			resultSet.close();
			
			statement.close();
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Method returns currently stored weights of given
	 * search engines for given query and algorithm.
	 * 
	 * @param query query to get the weights for
	 * @param algorithm algorithm to get the weights for
	 * @return Map of search engine ids with associated weights
	 */
	public Map<Integer, Float> getCurrentWeights(String query, String algorithm)
	{
		Map<Integer, Float> weights = new HashMap<Integer, Float>();
		
		try
		{
			String sql = "SELECT search_engine_id, weight FROM query_data WHERE query = ?" +
			" AND algorithm_id = (SELECT id FROM algorithm WHERE name = ?)";
	
			PreparedStatement statement = connection.prepareStatement(sql);
			
			statement.setString(1, query);
			statement.setString(2, algorithm);
	
			ResultSet resultSet = statement.executeQuery();
			
			while(resultSet.next())
			{
				weights.put(resultSet.getInt("search_engine_id"), resultSet.getFloat("weight"));
			}
			
			statement.close();
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return weights;
	}
}
