package pw.stadnik.searcher.agent;

import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.CyclicBehaviour;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.wrapper.AgentContainer;
import jade.wrapper.AgentController;
import jade.wrapper.StaleProxyException;

import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import pw.stadnik.common.HTMLTagA;
import pw.stadnik.common.SearchParams;
import pw.stadnik.searcher.algorithms.AlgorithmHelper;
import pw.stadnik.searcher.algorithms.AuctionMethod;
import pw.stadnik.searcher.algorithms.ConsensusMethod;
import pw.stadnik.searcher.algorithms.GameTheoryMethod;
import pw.stadnik.searcher.algorithms.MethodType;
import pw.stadnik.searcher.searchengine.SearchEngine;
import pw.stadnik.searcher.searchengine.SearchEngineDataAccess;

/**
 * 
 * Manager agent is a main class of the application.
 * Essentially it controls the flow of information
 * throughout the whole system. It creates every
 * single search agent that is to be utilized
 * in the system. It processes the whole data that
 * is being returned by the search agents. It controls
 * the algorithms for answer processing. It is the only
 * way of communication between client and main modules.
 * 
 * @author Jakub Stadnik
 *
 */
public class ManagerAgent extends Agent
{

	/**
	 * Unique random identifier assigned during creation process.
	 */
	private String uniqueId;
	/**
	 * ManagerAgent name.
	 */
	private String name;
	/**
	 * List of controllers of SearchAgents.
	 */
	private List<AgentController> agentControllers = new ArrayList<AgentController>();
	/**
	 * Serial version UID. It is used when 
	 * serializing a particular object of this class.
	 */
	private static final long serialVersionUID = 8840147715823767833L;
	/**
	 * Initial list of configurations of search agents.
	 * It does not contain any references to given
	 * agent objects/
	 */
	private List<AID> searchAgents;
//	private ArrayList<AID> searchAgentsIds;
	/**
	 * List of search engines. It does contain
	 * every search engine as an object which
	 * in turn can be sent to a search agent on
	 * its request.
	 */
	private List<SearchEngine> searchEngines;
	/**
	 * Map of currently taken engines and 
	 * agent identifiers corresponding to given engine.
	 */
	private Map<Integer, AID> takenEngines;
	/**
	 * Map of results of search agent identifiers 
	 * after they perform search process.
	 */
	private Map<AID, List<HTMLTagA>> obtainedResults = new HashMap<AID, List<HTMLTagA>>();
	/**
	 * Selected method of answer processing
	 * algorithm.
	 */
	private MethodType selectedMethod;// = MethodType.GAMETHEORY;
	/**
	 * Query that is sent initially by the user.
	 */
	private List<String> phrases;
	/**
	 * Identifier of gateway agent. This agent
	 * is the one that the initial query is coming from.
	 * It is also the one who receives answers.
	 */
	private AID gatewayAgent;
	/**
	 * Consensus answer used when selected method
	 * is the consensus method.
	 */
	private List<HTMLTagA> consensus = new ArrayList<HTMLTagA>();
	/**
	 * Agent identifiers which will receive feedback after
	 * it is received from user.
	 */
	private List<AID> feedbackAgents = new ArrayList<AID>();
	/**
	 * Map of agent identifiers and their corresponding
	 * weights.
	 */
	private Map<AID, Float> weights = null;
	/**
	 * List of results that is sent to the user 
	 * to receive feedback.
	 */
	private List<HTMLTagA> resultsForFeedback = new ArrayList<HTMLTagA>();
	/**
	 * Ranking of results. Used when selected methods
	 * are Auction and Game Theory.
	 */
	private Map<AID, List<HTMLTagA>> linksRanking = null;
	/**
	 * Feedback that is received from the user.
	 */
	private HTMLTagA returnedFeedback = null;
	private String fileName;
	private String joinedQuery;
	/**
	 * Method clears fields for the next set of queries.
	 */
	private void clearFields()
	{
		//this.searchAgents.clear();
		//this.searchEngines.clear();
		//this.takenEngines.clear();
		//this.confirmedAgents.clear();
		this.obtainedResults.clear();
		this.phrases.clear();
		this.consensus.clear();
		this.feedbackAgents.clear();
		//this.agentControllers.clear();
	}
	
	/* (non-Javadoc)
	 * @see jade.core.Agent#setup()
	 */
	public void setup()
	{
		Object[] args = getArguments();
		this.name = (String)args[0];
		System.out.println("Agent " + getLocalName() + " started");
		getSearchEngines();
		setupSearchAgents();
		propagateEngines();
		addBehaviour(new CyclicBehaviour(this)
		{
			/**
			 * 
			 */
			private static final long serialVersionUID = 6237029653653881282L;
//			private boolean stop = false;
			@Override
			public void action()
			{	
				processQuery(false);
			}
			
			public int onEnd() 
			{
				myAgent.doDelete();
				return super.onEnd();
		    } 
		});
	}
	
	/**
	 * Method assigns the search engines to given
	 * search engines. It sends the SearchEngine objects
	 * to agents.
	 */
	private void propagateEngines()
	{
		if(takenEngines == null)
			takenEngines = new HashMap<Integer, AID>();
		for(int i = 0; i < searchEngines.size(); i++)
		{
			try
			{
				ACLMessage engineMessage = new ACLMessage(ACLMessage.PROPAGATE);
				engineMessage.setContentObject(searchEngines.get(i));
				engineMessage.addReceiver(searchAgents.get(i));
				send(engineMessage);
				takenEngines.put(searchEngines.get(i).getId(), searchAgents.get(i));
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
		}
		
	}

	/** 
	 * Method is invoked when agent is being killed.
	 * It is overridden to clean up container from the SearchAgents
	 * associated with the ManagerAgent. 
	 **/
	public void doDelete()
	{
		for(int i = 0; i < agentControllers.size(); i++)
		{
			try
			{
				agentControllers.get(i).kill();
			}
			catch (StaleProxyException e)
			{
				e.printStackTrace();
			}
		}
		super.doDelete();
	}
	private Float findSetCoverage(List<HTMLTagA> list1, List<HTMLTagA> list2)
	{
		Float setCoverage = 0.0f;
		
		for(int i = 0; i < 10; i++)
		{
			if(list2.contains(list1.get(i)))
				setCoverage += 1.0f;
		}
		
		setCoverage /= 10.0f;
		
		return setCoverage;
	}
	
	private Float findURLCoverage(List<HTMLTagA> list1, List<HTMLTagA> list2)
	{
		Float URLCoverage = 0.0f;
		
		for(int i = 0; i < 10; i++)
		{
			if(list2.get(i).equals(list1.get(i)))
				URLCoverage += 1.0f;
		}
		
		URLCoverage /= 10.0f;
		
		return URLCoverage;
	}
	/**
	 * Method wraps all necessary methods to perform
	 * query processing.
	 * 
	 * @param searchParamsPresent true if initial search parameters are present
	 * 		false means that agent will wait for search parameters submission
	 */
	private void processQuery(boolean searchParamsPresent)
	{
		if(!searchParamsPresent)
			getSearchParams();
		//getSearchEngines();
		//setupSearchAgents();
		broadcastQuery();
		broadcastAlgorithm();
		//negotiateEngines();
		getResults();
		if(!(selectedMethod == MethodType.STATISTICS))
		{
			Map<AID, Float> currentWeights = getCurrentWeights();
			List<HTMLTagA> answer = performAnswerProcessing(currentWeights);
			if(answer == null)
			{
				try
				{
					
					
					BufferedWriter bw = new BufferedWriter(
							new OutputStreamWriter(
							new FileOutputStream("c:\\" + fileName +".txt", true)));
					
					bw.write(this.selectedMethod.toString() + "\tfeedback required\t" +
							new Timestamp(Calendar.getInstance().getTimeInMillis()).toString() + "\r\n");
					
					bw.write("\r\n");
					
					for(int i = 0; i < 10; i++)
					{
						HTMLTagA tagA = resultsForFeedback.get(i);
						bw.write(tagA.toString() + ";" + tagA.getEngineName() + "\r\n");
					}
					bw.close();
				}
				catch(Exception e)
				{
					e.printStackTrace();
				}
				sendAnswer(resultsForFeedback, true);
				boolean feedbackReceived = receiveFeedback();
				if(feedbackReceived)
				{
					calculateWeights();
					updateKnowledgeBase();
			
				}
			}
			else
			{
				try
				{
					
					
					BufferedWriter bw = new BufferedWriter(
							new OutputStreamWriter(
							new FileOutputStream("c:\\" + fileName +".txt", true)));
					
					bw.write(this.selectedMethod.toString() + "\t" + 
							new Timestamp(Calendar.getInstance().getTimeInMillis()).toString() + "\r\n");
					
					bw.write("\r\n");
					
					for(int i = 0; i < answer.size(); i++)
					{
						HTMLTagA tagA = answer.get(i);
						bw.write(tagA.toString() + ";" + tagA.getEngineName() + "\r\n");
					}
					bw.close();
				}
				catch(Exception e)
				{
					e.printStackTrace();
				}
				sendAnswer(answer, false);
				updateKnowledgeBase();
			}
		}
		/*else
		{
			String fileName = this.generateStatistics();
			sendFileName(fileName);
		}*/
		clearFields();
	}
	
	/**
	 * Method broadcasts weights to agents
	 * that are to update their knowledge bases.
	 * To each agent its corresponding weight
	 * is sent.
	 */
	private void updateKnowledgeBase()
	{
		for(AID sa : weights.keySet())
		{
			ACLMessage weight = new ACLMessage(ACLMessage.INFORM);
			try
			{
				weight.setContentObject(weights.get(sa));
			}
			catch (Exception e) 
			{
				e.printStackTrace();
			}
			weight.addReceiver(sa);
			send(weight);
		}
	}
	
	/**
	 * Method puts agent in waiting state.
	 * Agent waits for the feedback from the user.
	 */
	private boolean receiveFeedback()
	{
		boolean feedbackReceived = true;
		try
		{
			ACLMessage message = blockingReceive();
			Object contentObject = message.getContentObject();
			if(contentObject instanceof HTMLTagA)
			{
				HTMLTagA tagA = (HTMLTagA)contentObject;
				this.returnedFeedback = tagA;
			}
			else if(contentObject instanceof SearchParams)
			{
				clearFields();
				SearchParams sp = (SearchParams)contentObject;
				this.phrases = sp.getPhrases();
				this.selectedMethod = MethodType.valueOf(sp.getAlgorithm());
				this.gatewayAgent = message.getSender();
				feedbackReceived = false;
				processQuery(true);
			}
			
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return feedbackReceived;
	}
	
	/**
	 * Method calculates weights after feedback
	 * is received. Depending on algorithm
	 * there are two different ways to calculate those.
	 */
	private void calculateWeights()
	{
		this.weights = new HashMap<AID, Float>();
		AID winner = null;
		if(selectedMethod == MethodType.CONSENSUS)
		{
			int minDistance = Integer.MAX_VALUE;
			for(AID sa : obtainedResults.keySet())
			{
				List<HTMLTagA> links = obtainedResults.get(sa);
				int indexOfFeedback = consensus.indexOf(returnedFeedback);
				int distance = 0;
				if(links.contains(returnedFeedback))
				{
					distance = Math.abs(links.indexOf(returnedFeedback) - indexOfFeedback);
					if(minDistance > distance)
					{
						minDistance = distance;
						winner = sa;
					}
				}
			}
			
			weights.put(winner, 1f);
			
			for(AID sa : obtainedResults.keySet())
			{
				if(sa == winner)
					continue;
				List<HTMLTagA> links = obtainedResults.get(sa);
				int differentLinks = 0;
				differentLinks = AlgorithmHelper.levensteinDistance(links, obtainedResults.get(winner));
				weights.put(sa, 1.0f - (float)differentLinks / Math.min((float)consensus.size(), (float)links.size()));
			}
		}
		else
		{
			float maxRank = Float.MIN_VALUE;
			for(AID sa : linksRanking.keySet())
			{
				float rank = linksRanking.get(sa).get(linksRanking.get(sa).indexOf(returnedFeedback)).getRank();
				if(rank > maxRank)
				{
					maxRank = rank;
					winner = sa;
				}
			}
			
			weights.put(winner, 1f);
			
			for(AID sa : linksRanking.keySet())
			{
				if(sa == winner)
					continue;
				int differentLinks = 0;
				for(HTMLTagA tagA : linksRanking.get(sa))
				{
					if(tagA.getRank() != linksRanking.get(winner).get(linksRanking.get(winner).indexOf((tagA))).getRank())
						differentLinks++;
				}
				weights.put(sa, (float) differentLinks / (float) linksRanking.get(sa).size());
			}
		}
	}
	
	/*private void sendFileName(String fileName)
	{
		ACLMessage message = new ACLMessage(ACLMessage.INFORM);
		try
		{
			message.addReceiver(gatewayAgent);
			message.setContentObject(fileName);
			send(message);
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}*/
	
	/**
	 * Method sends answer to the feedback agent.
	 * 
	 * @param answer answer that is yield by the system
	 */
	private void sendAnswer(List<HTMLTagA> answer, Boolean feedbackRequired)
	{
		ACLMessage message = new ACLMessage(ACLMessage.INFORM);
		try
		{
			message.addReceiver(gatewayAgent);
			message.setContentObject((ArrayList<HTMLTagA>)answer);
			send(message);
			message = new ACLMessage(ACLMessage.INFORM);
			message.addReceiver(gatewayAgent);
			message.setContentObject(feedbackRequired);
			send(message);
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Method puts agent into the waiting state.
	 * It waits for the initial search parameters
	 * to be given by the user.
	 */
	private void getSearchParams()
	{
		ACLMessage message = blockingReceive(MessageTemplate.MatchPerformative(ACLMessage.REQUEST));
		try
		{
			SearchParams sp = (SearchParams)message.getContentObject();
			this.phrases = sp.getPhrases();
			this.selectedMethod = MethodType.valueOf(sp.getAlgorithm());
			this.gatewayAgent = message.getSender();
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Method broadcasts the selected algorithm
	 * to search agents.
	 */
	private void broadcastAlgorithm()
	{
		ACLMessage message = new ACLMessage(ACLMessage.INFORM);
		message.setContent(this.selectedMethod.toString());
		for(AID agent : searchAgents)
		{
			message.addReceiver(agent);
		}
		send(message);
		
	}
	
	/**
	 * Method performs answer processing.
	 * Depending on selected algorithm it invokes
	 * a particular method.
	 * 
	 * @param currentWeights currently assigned weights
	 * 
	 * @return answer already processed by the system
	 */
	private List<HTMLTagA> performAnswerProcessing(Map<AID, Float> currentWeights)
	{
		switch (this.selectedMethod)
		{
			case AUCTION:
				return auctionMethod(currentWeights);
			case CONSENSUS:
				return consensusMethod(currentWeights);
			case GAMETHEORY:
				return gameTheoryMethod(currentWeights);
			case STATISTICS:
				return null;
			default:
				return null;
		}
	}
	
	/**
	 * Method invokes statistics generation.
	 * 
	 * @return name of the file containing statistics
	 */
	/*private String generateStatistics()
	{
		Map<String, List<HTMLTagA>> results = new HashMap<String, List<HTMLTagA>>();
		for(AID aid : obtainedResults.keySet())
		{
			for(Integer engineId : takenEngines.keySet())
			{
				if(aid.toString().compareToIgnoreCase(takenEngines.get(engineId).toString()) == 0)
				{
					for(int i = 0; i < searchEngines.size(); i++)
					{
						if(engineId == searchEngines.get(i).getId())
						{
							SearchEngine engine = searchEngines.get(i);
							
							results.put(engine.getName(), obtainedResults.get(aid));
						}
					}
				}
			}
		}
		
		String query = phrases.get(0);
		
		for(int i = 1; i < phrases.size(); i++)
		{
			query += " " + phrases.get(i);
		}
		
		String fileName = "Statistics for " + query + ".xls";
		
		try
		{
			Statistics statistics = new Statistics();
			statistics.generateReport(new FileOutputStream(fileName), query, results);
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return fileName;
		
	}*/
	
	/**
	 * Method constructs result set that is
	 * to be sent to the user for feedback.
	 */
	private void constructFeedback()
	{
		resultsForFeedback.clear();
		if(selectedMethod == MethodType.CONSENSUS)
		{
			resultsForFeedback.addAll(this.consensus);
		}
		else
		{
			for(AID sa : linksRanking.keySet())
			{
				for(HTMLTagA tagA : linksRanking.get(sa))
				{
					if(!resultsForFeedback.contains(tagA))
						resultsForFeedback.add(tagA);
				}
			}
		}
	}
	
	/**
	 * Consensus method flow.
	 * 
	 * @param currentWeights currently assigned weights
	 * 
	 * @return answer list yield by this method
	 */
	private List<HTMLTagA> consensusMethod(Map<AID, Float> currentWeights)
	{
		ConsensusMethod cm = new ConsensusMethod();
		cm.findConsensusAnswer(obtainedResults, currentWeights);
		if(!cm.determineConsistence())
		{
			this.consensus = cm.getConsensus();
			constructFeedback();
			return null;
		}
		weights = cm.calculateWeights();
		List<HTMLTagA> answer = cm.getWinnerLinks();
		return answer;
		
	}
	/**
	 * Game Theory method flow.
	 * 
	 * @param currentWeights currently assigned weights
	 * 
	 * @return answer list yield by this method
	 */
	@SuppressWarnings("unchecked")
	private List<HTMLTagA> gameTheoryMethod(Map<AID, Float> currentWeights)
	{
		//AID loser = null;
		GameTheoryMethod gtm = new GameTheoryMethod();
		if(!gtm.filterAndRankResults(obtainedResults, currentWeights)) //cannot perform algorithm
		{
			linksRanking = gtm.getRanking();
			constructFeedback();
			return null;
		}
		List<HTMLTagA> answer = new ArrayList<HTMLTagA>();
		
		@SuppressWarnings("unused")
		Map<AID, List<HTMLTagA>> ranking = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)gtm.getRanking()).clone();
		
		for(int i = 0; i < 10; i++)
		{
			System.out.println("Answer number " + i);
			while(gtm.getAgentCount() > 1)
			{
				gtm.nextRound();
			}
			if(i == 0)
				weights = gtm.calculateWeights();
			HTMLTagA winnerLink = gtm.getWinnerLink();
			answer.add(winnerLink);
			
			for(AID a : ranking.keySet())
			{
				ranking.get(a).remove(winnerLink);
			}
			gtm.updateRanking(ranking);
		}
		return answer;
	}
	/**
	 * Auction method flow.
	 * 
	 * @param currentWeights currently assigned weights
	 * 
	 * @return answer list yield by this method
	 */
	@SuppressWarnings("unchecked")
	private List<HTMLTagA> auctionMethod(Map<AID, Float> currentWeights)
	{
		AuctionMethod am = new AuctionMethod();
		if(am.filterAndRankResults(obtainedResults, currentWeights) == false)
		{
			linksRanking = am.getRanking();
			constructFeedback();
			return null;
		}
		List<HTMLTagA> answer = new ArrayList<HTMLTagA>();
		
		@SuppressWarnings("unused")
		Map<AID, List<HTMLTagA>> ranking = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)am.getRanking()).clone();
		
		for(int i = 0; i < 10; i++)
		{
			System.out.println("Answer number " + i);
			while(am.getAgentCount() > 1)
			{
				am.nextRound();
			}
			if(i == 0)
				weights = am.calculateWeights();
			HTMLTagA winnerLink = am.getWinnerLink();
			answer.add(winnerLink);
			
			for(AID a : ranking.keySet())
			{
				ranking.get(a).remove(winnerLink);
			}
			
			am.updateRanking(ranking);
		}
		return answer;
	}
	
	/**
	 * This method is a process of negotiation
	 * of search engines with search agents.
	 * Search agents query the manager about the
	 * possibility of using the requested search engine.
	 * Manager decides which engine is to be given to which
	 * agent depending on the order of requests.
	 */
	/*private void negotiateEngines()
	{
		takenEngines = new HashMap<Integer, AID>();
		ArrayList<AID> erroredAgents = new ArrayList<AID>();
		while(takenEngines.size() < searchAgents.size())
		{
			ACLMessage searcherRequest = blockingReceive(MessageTemplate.MatchPerformative(ACLMessage.REQUEST));
			if(searcherRequest != null)
			{	
				AID searcher = searcherRequest.getSender();
				Integer searchEngineId = Integer.parseInt(searcherRequest.getContent());
				SearchEngine se = null;
				
				if(takenEngines.containsKey(searchEngineId))
				{
					ACLMessage refusal = new ACLMessage(ACLMessage.REFUSE);
					refusal.addReceiver(searcher);
					send(refusal);
					continue;
				}
				
				for(int i = 0; i < searchEngines.size(); i++)
				{
					if(searchEngines.get(i).getId() == searchEngineId)
						se = searchEngines.get(i);
				}
				
				ACLMessage agreement = new ACLMessage(ACLMessage.AGREE);
				agreement.addReceiver(searcher);
				try
				{
					agreement.setContentObject(se);
				}
				catch(Exception e)
				{
					e.printStackTrace();
				}
				takenEngines.put(searchEngineId, searcher);
				send(agreement);
			}
		}
		while(confirmedAgents.size() < takenEngines.size())
		{
			ACLMessage searcherConfirmation = blockingReceive(MessageTemplate.MatchPerformative(ACLMessage.INFORM));
			if(searcherConfirmation != null)
			{
				if(searcherConfirmation.getContent().equals("OK"))
				{
					confirmedAgents.add(searcherConfirmation.getSender());
					System.out.println("Agent confirmed: " + searcherConfirmation.getSender().getLocalName());
				}
				else if(searcherConfirmation.getContent().equals("NOT OK"))
				{
					erroredAgents.add(searcherConfirmation.getSender());
				}
			}
		}
		//confirmedAgents.clear();
		//takenEngines.clear();
	}*/
	
	/**
	 * Method retrieves search engines from
	 * the database. It also retrieves engines
	 * configurations.
	 */
	private void getSearchEngines()
	{
		SearchEngineDataAccess dataAccess = new SearchEngineDataAccess();
		dataAccess.openConnection();
		
		searchEngines = dataAccess.getEngines();
		
		dataAccess.closeConnection();
		
		for(int i = 0; i < searchEngines.size(); i++)
		{
			searchEngines.get(i).getIgnoreData();
		}
	}
	
	/**
	 * Method broadcasts query to 
	 * search agents.
	 */
	private void broadcastQuery()
	{
		ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
		msg.setContent("Query");
		try
		{
			msg.setContentObject((ArrayList<String>)phrases);
		
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		ACLMessage resultCountMessage = new ACLMessage(ACLMessage.INFORM);
		resultCountMessage.setContent("20");
		for(AID agent : searchAgents)
		{
			msg.addReceiver(agent);
			resultCountMessage.addReceiver(agent);
		}
		send(msg);
		send(resultCountMessage);
	}
	
	/**
	 * Method puts agent into the waiting state.
	 * It waits for search results that are to 
	 * be received from search agents.
	 */
	@SuppressWarnings("unchecked")
	private void getResults()
	{
		while(obtainedResults.keySet().size() < takenEngines.size())
		{
			ACLMessage message = blockingReceive(MessageTemplate.MatchPerformative(ACLMessage.PROPOSE));
			try
			{
				AID sender = message.getSender();
				ArrayList<HTMLTagA> results = (ArrayList<HTMLTagA>)message.getContentObject();
				obtainedResults.put(sender, results);
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
		}
		
		fileName = this.selectedMethod.toString();
		
		joinedQuery = this.phrases.get(0);
		
		for(int i = 1; i < phrases.size(); i++)
		{
			joinedQuery += " " + phrases.get(i);
		}
		
		fileName += "_" + joinedQuery;
		
		try
		{
			BufferedWriter bw = new BufferedWriter(
					new OutputStreamWriter(
					new FileOutputStream("c:\\" + fileName +".txt")));
			
			for(AID agent : obtainedResults.keySet())
			{
				List<HTMLTagA> results = obtainedResults.get(agent);
				
				String engineName = results.get(0).getEngineName();
				
				bw.write(engineName.toUpperCase() + "\r\n");
				
				bw.write("\r\n");
				
				for(int i = 0; i < Math.min(10, results.size()); i++)
				{
					bw.write(results.get(i).toString());
					bw.write("\r\n");
				}
				
				bw.write("\r\n");
				bw.write("\r\n");
			}
			
			bw.close();
		}
		catch(Exception e)
		{
			
		}
	}
	
	/**
	 * Method creates and setups new search agents.
	 * Number of created search agents depends on the
	 * number of configured search engines. The newly
	 * created agents are immediately started.
	 */
	private void setupSearchAgents()
	{
		AgentContainer container = getContainerController();
		
		for(int i = 0; i < searchEngines.size(); i++)
		{
			try
			{
				if(searchAgents == null)
					searchAgents = new ArrayList<AID>();
				this.uniqueId = this.name.split("_")[1];
				Object[] args = new Object[2];
				args[0] = i + 1;
				args[1] = "SearchAgent" + String.valueOf(i + 1) + "_" + uniqueId;
				AgentController a = container.createNewAgent(String.valueOf(args[1]), "pw.stadnik.searcher.agent.SearchAgent", args);
				a.start();
				agentControllers.add(a);
				AID searchAgentAID = new AID(String.valueOf(args[1]), AID.ISLOCALNAME);
				searchAgents.add(searchAgentAID);
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
		}
	}

	/**
	 * Method returns weights associated with the given
	 * search agents.
	 * 
	 * @return Map of agents with their associated weights
	 */
	private Map<AID, Float> getCurrentWeights()
	{
		Map<AID, Float> weights = new HashMap<AID, Float>();
		
		AgentDataAccess dataAccess = new AgentDataAccess();
		dataAccess.openConnection();
		String query = "";
		query = phrases.get(0);
		for(int i = 1; i < phrases.size(); i++)
		{
			query = query + " " + phrases.get(i);
		}
		
		Map<Integer, Float> currentWeights = dataAccess.getCurrentWeights(query, this.selectedMethod.toString());
		
		for(Integer searchEngineId : takenEngines.keySet())
		{
			if(currentWeights.containsKey(searchEngineId))
			{
				weights.put(takenEngines.get(searchEngineId), currentWeights.get(searchEngineId));
			}
			else
			{
				weights.put(takenEngines.get(searchEngineId), 1.0f);
			}
		}
		dataAccess.closeConnection();
		
		return weights;
	}
}
