package pw.stadnik.searcher.algorithms;

import jade.core.AID;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import pw.stadnik.common.HTMLTagA;

/**
 * Class provides access to common
 * methods used in algorithms.
 * 
 * @author Jakub Stadnik
 *
 */
public class AlgorithmHelper
{
	/**
	 * Method takes as parameter map of results obtained
	 * from search agents after the search was performed.
	 * The second parameter is an out parameter used in
	 * Game Theory and Auction Method. Method ranks answers 
	 * and updates result sets. Method can return <code>null</code>
	 * if the initial result sets are adjacent (have no common answer)
	 * or full ranking of the results. The out parameter is 
	 * basically the same as returned result, however it will always be a ranking
	 * no matter if the sets were adjacent or not.
	 * 
	 * @param results initial search results
	 * @param finalLinksRanking (out) full links ranking
	 * @param currentWeights currently assigned weights
	 * @return full links ranking or <code>null</code>
	 */
	@SuppressWarnings("unchecked")
	public static Map<AID, List<HTMLTagA>> filterAndRankResults(Map<AID, List<HTMLTagA>> results, Map<AID, List<HTMLTagA>> finalLinksRanking, Map<AID, Float> currentWeights)
	{
		Map<AID, List<HTMLTagA>> linksRanking = new HashMap<AID, List<HTMLTagA>>();//new HashMap<AID, Map<HTMLTagA,Float>>();
		//this.initialResults = (HashMap<AID, ArrayList<HTMLTagA>>)results.clone();
		//this.agents = results.keySet();
		Map<AID, Map<AID, Integer>> answersMatrix = new HashMap<AID, Map<AID,Integer>>();
		Set<AID> tempSet = (Set<AID>)((HashMap<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)results).clone()).keySet();
		for(AID agent : results.keySet())
		{
			for(AID secondAgent : tempSet)
			{
				if(agent.equals(secondAgent))
					continue;
				List<HTMLTagA> links = results.get(agent);
				List<HTMLTagA> secondAgentLinks = results.get(secondAgent);
				Integer sameResults = 0;
				for(int i = 0; i < links.size(); i++)
				{
					if(secondAgentLinks.contains(links.get(i)))
					{
						sameResults++;
					}
				}
				if(!answersMatrix.containsKey(agent))
				{
					Map<AID, Integer> secondAgentSameResults = new HashMap<AID, Integer>();
					secondAgentSameResults.put(secondAgent, sameResults);
					answersMatrix.put(agent, secondAgentSameResults);
				}
				else
				{
					Map<AID, Integer> secondAgentSameResults = answersMatrix.get(agent);
					secondAgentSameResults.put(secondAgent, sameResults);
					answersMatrix.put(agent, secondAgentSameResults);
				}
			}
		}
		ArrayList<AID> emptyAgents = new ArrayList<AID>();
		for(AID agent : results.keySet())
		{
			Integer numberOfZeros = 0;
			for(AID secondAgent : tempSet)
			{
				if(agent.equals(secondAgent))
					continue;
				if(answersMatrix.get(agent).get(secondAgent) == 0)
					numberOfZeros++;
			}
			//no common answers with any agent
			if(numberOfZeros == results.keySet().size() - 1)
				emptyAgents.add(agent);
		}
		
		for(int i = 0; i < emptyAgents.size(); i++)
			results.remove(emptyAgents.get(i));
		for(AID agent : results.keySet())
		{
			List<HTMLTagA> links = results.get(agent);
			Float weight = currentWeights.get(agent);
			List<HTMLTagA> ranking = new ArrayList<HTMLTagA>();
			for(int i = 0; i < links.size(); i++)
			{
				HTMLTagA link = (HTMLTagA)links.get(i).clone();
				
				link.setRank((float)(links.size() - i) * weight);
				ranking.add(link);
			}
			linksRanking.put(agent, ranking);
		}
		for(AID agent : linksRanking.keySet())
		{
			List<HTMLTagA> links = linksRanking.get(agent);
			for(int i = 0; i < links.size(); i++)
			{
				HTMLTagA link = links.get(i);
				List<AID> agentsForUpdate = findAgentsWithMissingLink(linksRanking, agent, link);
				
				for(int j = 0; j < agentsForUpdate.size(); j++)
				{
					List<HTMLTagA> updatedRanking = linksRanking.get(agentsForUpdate.get(j));
					Float weight = currentWeights.get(agentsForUpdate.get(j));
					HTMLTagA clonedLink = (HTMLTagA)link.clone();
					if(weight == 0)
						weight = 0.01f;
					clonedLink.setRank(1.0f * weight);
					updatedRanking.add(clonedLink);
					linksRanking.put(agentsForUpdate.get(j), updatedRanking);
				}
			}
		}
		//checkIntegrity(linksRanking);
		for(AID a : linksRanking.keySet())
		{
			List<HTMLTagA> links = new ArrayList<HTMLTagA>();//, Float>();
			for(HTMLTagA tagA : linksRanking.get(a))
			{
				links.add((HTMLTagA)tagA.clone());
			}
			finalLinksRanking.put(a, links);
		}
		if(emptyAgents.size() == results.keySet().size())
			return null;
		return linksRanking;
	}
	
	/*private static void checkIntegrity(Map<AID, List<HTMLTagA>> linksRanking)
	{
		for(AID a : linksRanking.keySet())
		{
			for(AID b : linksRanking.keySet())
			{
				if(a.equals(b))
					continue;
				List<HTMLTagA> listA = linksRanking.get(a);
				List<HTMLTagA> listB = linksRanking.get(b);
				
				for(int i = 0; i < listA.size(); i++)
				{
					if(!listB.contains(listA.get(i)))
						System.out.println(listA.get(i));
				}
				
				for(int i = 0; i < listB.size(); i++)
				{
					if(!listA.contains(listB.get(i)))
						System.out.println(listB.get(i));
				}
			}
		}
	}*/
	
	/**
	 * Method finds agents which results sets
	 * are missing a link from a result set
	 * of a particular agent.
	 * 
	 * @param results map of results of all agents
	 * @param agent agent from which result set the link is checked
	 * @param link link to find in the results
	 * @return list of agents which are missing the link
	 */
	private static List<AID> findAgentsWithMissingLink(Map<AID, List<HTMLTagA>> results, AID agent, HTMLTagA link)
	{
		List<AID> agents = new ArrayList<AID>();
		for(AID aid : results.keySet())
		{
			if(aid.equals(agent))
				continue;
			if(!results.get(aid).contains(link))
			{
				agents.add(aid);
			}
		}
		return agents;
	}
	
	/**
	 * Method calculates weights for Game Theory
	 * and Auction method
	 * 
	 * @param linksRanking full ranking of links
	 * @param initialResults map of initial results
	 * @return map of agents with corresponding weights
	 */
	public static Map<AID, Float> calculateWeights(Map<AID, List<HTMLTagA>> linksRanking, Map<AID, List<HTMLTagA>> initialResults)
	{
		Map<AID, Float> weights = new HashMap<AID, Float>();
		AID winner = null;
		
		if(!(linksRanking.keySet().size() == 1))
			return null;
		
		for(AID sa : linksRanking.keySet())
			winner = sa;
		
		weights.put(winner, 1f);
		
		for(AID sa : initialResults.keySet())
		{
			if(sa == winner)
				continue;
			int differentLinks = findDifferentLinks(initialResults.get(winner), initialResults.get(sa));
			
			weights.put(sa, ((float)initialResults.get(sa).size() - (float) differentLinks) / (float)initialResults.get(sa).size());
		}
		
		return weights;
	}
	
	/**
	 * Method calculates average of values
	 * of the link ranks
	 * 
	 * @param values ranks of the link
	 * @return average value of ranks
	 */
	public static Float calculateAverage(List<Float> values)
	{
		Float average = new Float(0.0f);
		for(int i = 0; i < values.size(); i++)
		{
			average += values.get(i);
		}
		average /= (float)values.size();
		return average;
	}
	
	/**
	 * Method finds the number of different links
	 * between two result sets. By different we mean 
	 * the link that is not contained in the second result set.
	 * 
	 * @param list1 first list of links
	 * @param list2 second list of links
	 * @return number of different links
	 */
	private static int findDifferentLinks(List<HTMLTagA> list1, List<HTMLTagA> list2)
	{
		int differentLinks = 0;
		
		for(HTMLTagA tagA : list2)
		{
			if(!list1.contains(tagA))
				differentLinks++;
		}
		
		return differentLinks;
	}
	
	/**
	 * Method calculates distance between
	 * two result sets.
	 * 
	 * @param links1 first list of links
	 * @param links2 second list of links
	 * @return distance between lists
	 */
	public static int findDistance(List<HTMLTagA> links1, List<HTMLTagA> links2)
	{
		int distance = 0;
		
		for(int i = 0; i < links1.size(); i++)
		{
			HTMLTagA link = links1.get(i);
			int index = links2.indexOf(link); 
			if(i != index)
			{
				if(index == -1)
				{
					distance += links1.size() + 1;
				}
				else
				{
					distance += Math.abs(i - index);
				}
			}
		}
		return distance;
	}
	
	/*public static int hausdorffDistance(List<HTMLTagA> list1, List<HTMLTagA> list2)
	{
		int distance = 0;
		
		List<Integer> d12 = new ArrayList<Integer>();
		List<Integer> d21 = new ArrayList<Integer>();
		
		int D12 = Integer.MIN_VALUE;
		int D21 = Integer.MIN_VALUE;
		
		
		for(int i = 0; i < list1.size(); i++)
		{
			int d = Integer.MAX_VALUE;
			HTMLTagA link1 = list1.get(i);
			for(int j = 0; j < list2.size(); j++)
			{
				HTMLTagA link2 = list2.get(j);
				int ld = levenshteinDistance(link1.getHREF().toCharArray(), link2.getHREF().toCharArray());
				if(d > ld)
				{
					d = ld;
				}
			}
			d12.add(d);
		}
		
		for(int i = 0; i < list2.size(); i++)
		{
			int d = Integer.MAX_VALUE;
			HTMLTagA link1 = list2.get(i);
			for(int j = 0; j < list1.size(); j++)
			{
				HTMLTagA link2 = list1.get(j);
				d = Math.min(d, levenshteinDistance(link1.getHREF().toCharArray(), link2.getHREF().toCharArray()));
			}
			d21.add(d);
		}
		
		for(int i = 0; i < d12.size(); i++)
		{
			D12 = Math.max(D12, d12.get(i));
		}
		
		for(int j = 0; j < d21.size(); j++)
		{
			D21 = Math.max(D12, d12.get(j));
		}
		
		distance = Math.max(D12, D21);
		
		return distance;
	}
	
	private static int levenshteinDistance(char[] str1, char[] str2)
	{	
		int[][] d = new int[str1.length + 1][str2.length + 1];
		
		int cost = 0;
		
		for(int i = 0; i <= str1.length; i++)
			d[i][0] = i;
		for(int j = 0; j <= str2.length; j++)
			d[0][j] = j;
		
		for(int i = 1; i <= str1.length; i++)
		{
			for(int j = 1; j <= str2.length; j++)
			{
				Character c1 = str1[i - 1];
				Character c2 = str2[j - 1];
				
				if(c1.equals(c2))
					cost = 0;
				else
					cost = 1;
				
				d[i][j] = Math.min(d[i - 1][j] + 1, Math.min(d[i][j - 1] + 1, d[i - 1][j - 1] + cost));
			}
		}
		
		return d[str1.length][str2.length];
	}*/
	
	/**
	 * Method calculates Levenshtein Distance
	 * between two lists.
	 * 
	 * @param list1 first list
	 * @param list2 second list
	 * @return distance
	 */
	public static int levensteinDistance(List<HTMLTagA> list1, List<HTMLTagA> list2)
	{
		int[][] d = new int[list1.size() + 1][list2.size() + 1];
		
		int cost = 0;
		
		for(int i = 0; i <= list1.size(); i++)
			d[i][0] = i;
		for(int j = 0; j <= list2.size(); j++)
			d[0][j] = j;
		
		for(int i = 1; i <= list1.size(); i++)
		{
			for(int j = 1; j <= list2.size(); j++)
			{
				HTMLTagA link1 = list1.get(i - 1);
				HTMLTagA link2 = list2.get(j - 1);
				if(link1.equals(link2))
					cost = 0;
				else
					cost = 1;
				
				d[i][j] = Math.min(d[i - 1][j] + 1, Math.min(d[i][j - 1] + 1, d[i - 1][j - 1] + cost));
			}
		}
		
		return d[list1.size()][list2.size()];
	}
}
