package pw.stadnik.searcher.algorithms;

import jade.core.AID;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import pw.stadnik.common.HTMLTagA;

/**
 * Class represents auction method algorithm.
 * It contains all necessary methods to 
 * perform the Auction method data processing.
 * 
 * @author Jakub Stadnik
 *
 */
public class AuctionMethod
{
	/**
	 * Map of costs of agents during
	 * auction process
	 */
	private Map<AID, Float> costs = new HashMap<AID, Float>();
	/**
	 * An agent that lost a round
	 */
	private AID loser;
	/**
	 * Full links ranking
	 */
	private Map<AID, List<HTMLTagA>> linksRanking = new HashMap<AID, List<HTMLTagA>>();
	/**
	 * Links that were chosen by agents
	 * in a round of negotiation process
	 */
	private Map<AID, HTMLTagA> agentsLinks = new HashMap<AID, HTMLTagA>();
	/**
	 * An agent that lost round before
	 */
	private AID latestLoser;
	/**
	 * Copy of full links ranking
	 */
	private Map<AID, List<HTMLTagA>> finalLinksRanking = null;
	/**
	 * Map of initial search results
	 */
	private Map<AID, List<HTMLTagA>> initialResults = null;
	
	/**
	 * Method updates links ranking to be used after removal
	 * of link which is already in the answer list.
	 * 
	 * @param ranking ranking after removal of link
	 */
	@SuppressWarnings("unchecked")
	public void updateRanking(Map<AID, List<HTMLTagA>> ranking)
	{
		this.linksRanking = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)ranking).clone();
	}
	
	/**
	 * Switch to tell if the situation with the same costs 
	 * of agents occurred twice.
	 */
	private boolean sameCostTwice = false;
	/**
	 * Method calculates cost for <code>agent</code>
	 * in each round of the negotiation.
	 * 
	 * @param agent an agent which cost should be calculated
	 * @param agentAnswerConfidence confidence of agent about his chosen answer
	 * @param contesterAnswerConfidences confidences of other agents about their chosen answers
	 */
	private void calculateCost(AID agent, Float agentAnswerConfidence, List<Float> contesterAnswerConfidences)
	{
		Float cost = 0.0f;
		for(int i = 0; i < contesterAnswerConfidences.size(); i++)
		{
			cost += agentAnswerConfidence - ((contesterAnswerConfidences.get(i) == null) ? 0.0f : contesterAnswerConfidences.get(i));
		}
		cost /= 10.0f;
		costs.put(agent, cost);
	}
	
	/**
	 * Returns full links ranking
	 * 
	 * @return links ranking for each agent
	 */
	public Map<AID, List<HTMLTagA>> getRanking()
	{
		return this.finalLinksRanking;
	}
	
	/**
	 * Finds the agent that won the negotiation
	 * 
	 * @return winner agent
	 */
	public AID getWinner()
	{
		if(linksRanking.keySet().size() == 1)
		{
			for(AID sa : linksRanking.keySet())
				return sa;
		}
		return null;
	}
	/**
	 * Returns number of agents that are
	 * still in the negotiation process
	 * 
	 * @return number of agents
	 */
	public Integer getAgentCount()
	{
		return this.linksRanking.keySet().size();
	}
	
	/**
	 * Method creates map of costs in each
	 * round of the negotiation
	 */
	private void createCostsMap()
	{
		costs.clear();
		for(AID agent : linksRanking.keySet())
		{
			List<Float> contesterAnswerConfidences = new ArrayList<Float>();
			for(AID contester : linksRanking.keySet())
			{
				if(agent.equals(contester))
					continue;
				HTMLTagA tagA = agentsLinks.get(contester);
				int index = linksRanking.get(agent).indexOf(tagA);
				float confidence = 1.0f;
				if(index != -1)
					confidence = linksRanking.get(agent).get(index).getRank();
				contesterAnswerConfidences.add(confidence);
			}
			calculateCost(agent, linksRanking.get(agent).get(linksRanking.get(agent).indexOf(agentsLinks.get(agent))).getRank(), contesterAnswerConfidences);
		}
	}
	
	/**
	 * Method invokes helper method from
	 * AlgorithmHelper class which in turn
	 * calculates weights.
	 * 
	 * @return map of weights calculated by helper
	 */
	public Map<AID, Float> calculateWeights()
	{
		return AlgorithmHelper.calculateWeights(linksRanking, initialResults);
	}
	
	/**
	 * Returns the link that won the negotiation
	 * 
	 * @return link which won the negotiation
	 */
	public HTMLTagA getWinnerLink()
	{
		if(linksRanking.keySet().size() == 1)
		{
			for(AID sa : linksRanking.keySet())
				return agentsLinks.get(sa);
		}
		return null;
	}
	
	/**
	 * Method determines which agent lost the round
	 * by finding agent with higest cost
	 */
	private void determineRoundLoser()
	{
		Float highestCost = Float.MIN_VALUE;
		
		for(AID sa : costs.keySet())
		{
			if(costs.get(sa) > highestCost)
			{
				highestCost = costs.get(sa);
				loser = sa;
			}
		}
		updateRankings();
		
		boolean sameCosts = true;
		
		Float lastCost = 0.0f;
		
		boolean isFirst = true;
		
		for(AID sa : costs.keySet())
		{
			if(isFirst)
			{
				lastCost = costs.get(sa);
				isFirst = false;
			}
			else
			{
				if(!lastCost.equals(costs.get(sa)))
					sameCosts = false;
			}
		}	
		
		if(sameCosts && !sameCostTwice)
		{
			sameCostTwice = true;
		}
		else if(sameCosts && sameCostTwice)
		{
			Float lowestRank = Float.MAX_VALUE;
			AID loser = null;
			for(AID sa : costs.keySet())
			{
				HTMLTagA tagA = finalLinksRanking.get(sa).get(finalLinksRanking.get(sa).
						indexOf(agentsLinks.get(sa)));
				if(tagA.getRank() < lowestRank)
				{
					lowestRank = tagA.getRank();
					loser = sa;
				}
			}
			
			linksRanking.remove(loser);
			costs.remove(loser);
			sameCostTwice = false;
			return;
		}
		
		/*if(linksRanking.size() == 2)
		{
			AID firstAgent = null;
			AID secondAgent = null;
			
			for(AID sa : costs.keySet())
			{
				if(firstAgent == null)
					firstAgent = sa;
				else
					secondAgent = sa;
			}
			
			Float firstCost = costs.get(firstAgent);
			Float secondCost = costs.get(secondAgent);
			
			if(firstCost.equals(secondCost) && !sameCostTwice)
			{
				sameCostTwice = true;
			}
			
			if(firstCost.equals(secondCost) && sameCostTwice)
			{
				sameCostTwice = false;
				
				
				HTMLTagA firstLink = finalLinksRanking.get(firstAgent).get(finalLinksRanking.get(firstAgent).
						indexOf(agentsLinks.get(firstAgent)));
				HTMLTagA secondLink = finalLinksRanking.get(secondAgent).get(finalLinksRanking.get(secondAgent).
						indexOf(agentsLinks.get(secondAgent)));
				
				if(firstLink.getRank() >= secondLink.getRank())
				{
					linksRanking.remove(secondAgent);
					costs.remove(secondAgent);
				}
				else
				{
					linksRanking.remove(firstAgent);
					costs.remove(secondAgent);
				}
				return;
			}
		}*/
		if(loser == latestLoser)
		{
			linksRanking.remove(latestLoser);
			costs.remove(latestLoser);
		}
		else
		{
			latestLoser = loser;
		}
	}
	
	/**
	 * Method finds the highest ranked 
	 * links for each agent
	 */
	private void findAgentsLinks()
	{
		agentsLinks.clear();
		for(AID agent : linksRanking.keySet())
		{
			Float confidence = Float.MIN_VALUE;
			List<HTMLTagA> agentLinks = linksRanking.get(agent);
			for(HTMLTagA tagA : agentLinks)
			{
				if(tagA.getRank() > confidence)
				{
					confidence = tagA.getRank();
					agentsLinks.put(agent, tagA);
				}
			}
		}
	}
	
	/**
	 * Method invokes next round of negotiation
	 */
	public void nextRound()
	{
		this.findAgentsLinks();
		this.createCostsMap();
		this.determineRoundLoser();
	}
	
	/**
	 * Method invokes the helper method from
	 * AlgorithmHelper class which in turn finds
	 * the links ranking.
	 * 
	 * @param results initial search results
	 * @param currentWeights currently assigned weights
	 * @return true if links were properly ranked (no disjoint result sets), false otherwise
	 */
	@SuppressWarnings("unchecked")
	public boolean filterAndRankResults(Map<AID, List<HTMLTagA>> results, Map<AID, Float> currentWeights)
	{	
		this.initialResults = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)results).clone();
		this.finalLinksRanking = new HashMap<AID, List<HTMLTagA>>();
		this.linksRanking = AlgorithmHelper.filterAndRankResults(results, this.finalLinksRanking, currentWeights);
		if(linksRanking == null)
			return false;
		return true;
	}
	
	/**
	 * Method updates rankings with new ranks
	 * after each round of negotiation
	 */
	private void updateRankings()
	{
		for(AID sa : this.costs.keySet())
		{
			HTMLTagA tagA = linksRanking.get(sa).get(linksRanking.get(sa).indexOf(agentsLinks.get(sa)));
			tagA.setRank(tagA.getRank() - costs.get(sa));
		}
	}
}
