package pw.stadnik.searcher.algorithms;

import jade.core.AID;

import java.util.*;

import pw.stadnik.common.HTMLTagA;

/**
 * Class represents consensus method algorithm.
 * It contains all necessary methods to 
 * perform the Consensus method data processing.
 * 
 * @author Jakub Stadnik
 *
 */
public class ConsensusMethod implements Comparator<HTMLTagA>
{
	/**
	 * Map containing initial search results
	 */
	private Map<AID, List<HTMLTagA>> initialResults = new HashMap<AID, List<HTMLTagA>>();
	/**
	 * Consensus answer
	 */
	private List<HTMLTagA> consensus = null;
	/**
	 * Map of weights
	 */
	private HashMap<AID, Float> weights = null;
	/**
	 * An agent that won the negotiation
	 */
	private AID winnerAgent = null;
	
	/**
	 * Method finds consensus answer.
	 * 
	 * @param results initial search results
	 * @param currentWeights currently assigned weights
	 * @return list of answers - consensus answer
	 */
	@SuppressWarnings("unchecked")
	public List<HTMLTagA> findConsensusAnswer(Map<AID, List<HTMLTagA>> results, Map<AID, Float> currentWeights)
	{
		this.consensus = new ArrayList<HTMLTagA>();
		this.initialResults = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)results).clone();
		List<HTMLTagA> documents = new ArrayList<HTMLTagA>();
		List<HTMLTagA> averages = new ArrayList<HTMLTagA>();//, Float>();
		for(AID sa : initialResults.keySet())
		{
			for(HTMLTagA tagA : initialResults.get(sa))
			{
				HTMLTagA clonedLink = (HTMLTagA)tagA.clone();
				if(!documents.contains(clonedLink))
					documents.add(clonedLink);
			}
		}
		
		for(HTMLTagA link : documents)
		{
			List<Float> positions = new ArrayList<Float>();
			for(AID sa : initialResults.keySet())
			{
				List<HTMLTagA> links = initialResults.get(sa);
				Float weight = currentWeights.get(sa);
				if(weight == 0.0f)
					weight = 0.01f;
				Integer index = links.indexOf(link);
				Float position = (float)index / weight;
				if(index == -1)
					position = (float)(links.size() + 1) / weight;
				positions.add(position);
			}
			Float average = AlgorithmHelper.calculateAverage(positions);
			link.setRank(average);
			averages.add(link);
		}
		
		Collections.sort(documents, this);
		
		for(int i = 0; i < documents.size(); i++)
		{
			consensus.add(documents.get(i));
		}
		
		return consensus;
	}
	
	/**
	 * Returns consensus answer
	 * 
	 * @return list of answers
	 */
	public List<HTMLTagA> getConsensus()
	{
		return this.consensus;
	}
	
	/**
	 * Method determines consistency of the consensus.
	 * 
	 * @return true if consensus answer is consistent, false otherwise
	 */
	@SuppressWarnings("unchecked")
	public boolean determineConsistence()
	{
		//find size of minimal non-zero result set
		
		int minSize = Integer.MAX_VALUE;
		
		Map<AID, List<HTMLTagA>> clonedData = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)initialResults).clone();
		
		for(AID sa : initialResults.keySet())
		{
			int resultSetSize = initialResults.get(sa).size();
			if(resultSetSize == 0)
			{
				clonedData.remove(sa);
				continue;
			}
			if(minSize > resultSetSize)
				minSize = resultSetSize;
		}
		
		initialResults = clonedData;
		//trim result sets and consensus
		
		//consensus = (ArrayList<HTMLTagA>)consensus.subList(0, minSize);
		
		List<HTMLTagA> tempList = (List<HTMLTagA>)((ArrayList<HTMLTagA>)consensus).clone();
		
		consensus.clear();
		
		for(int i = 0; i < minSize; i++)
		{
			consensus.add(tempList.get(i));
		}
		
		
		for(AID sa : initialResults.keySet())
		{
			List<HTMLTagA> links = initialResults.get(sa);
			tempList = (List<HTMLTagA>)((ArrayList<HTMLTagA>)links).clone();
			links.clear();
			for(int i = 0; i < minSize; i++)
			{
				links.add(tempList.get(i));
			}
			initialResults.put(sa, links);
		}
		
		//find average of distances
		
		float averageOfDistances = 0.0f;

		clonedData = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)initialResults).clone();
		
		for(AID sa : initialResults.keySet())
		{
			for(AID sa2 : clonedData.keySet())
			{
				if(sa != sa2)
				{
					List<HTMLTagA> saLinks = initialResults.get(sa);
					List<HTMLTagA> sa2Links = initialResults.get(sa2);
					averageOfDistances += AlgorithmHelper.levensteinDistance(saLinks, sa2Links);
				}
			}
		}
		
		averageOfDistances /= (float)(initialResults.keySet().size() * initialResults.keySet().size()); 
		
		float averageOfConsensusDistances = 0.0f;
		
		for(AID sa : initialResults.keySet())
		{
			averageOfConsensusDistances += AlgorithmHelper.levensteinDistance(initialResults.get(sa), consensus);
		}
		
		averageOfConsensusDistances /= (float)initialResults.keySet().size();
		
		return averageOfDistances > averageOfConsensusDistances;
	}
	
	/**
	 * Returns top ten links
	 * 
	 * @return first link from consensus answer
	 */
	public List<HTMLTagA> getWinnerLinks()
	{
		List<HTMLTagA> winnerLinks = new ArrayList<HTMLTagA>();
		for(int i = 0; i < 10; i++)
		{
			winnerLinks.add(consensus.get(i));
		}
		return winnerLinks;
	}
	
	/**
	 * Method calculates weights
	 * for Consensus method
	 * 
	 * @return map of weights
	 */
	public HashMap<AID, Float> calculateWeights()
	{
		int minDistance = Integer.MAX_VALUE;
		
		this.weights = new HashMap<AID, Float>();
		
		for(AID sa : initialResults.keySet())
		{
			int difference = AlgorithmHelper.levensteinDistance(initialResults.get(sa), consensus);
			if(minDistance > difference)
			{
				minDistance = difference;
				this.winnerAgent = sa;
			}
		}
		
		weights.put(winnerAgent, 1f);
		
		for(AID sa : initialResults.keySet())
		{
			if(sa == winnerAgent)
				continue;
			int difference = AlgorithmHelper.levensteinDistance(initialResults.get(sa), initialResults.get(winnerAgent));
			
			weights.put(sa, 1.0f - (float)difference / Math.min((float) initialResults.get(sa).size(), 
					(float)initialResults.get(winnerAgent).size()));
		}
		return weights;
	}
	
	/**
	 * Finds the agent whose answer set contains
	 * highest ranked link in the consensus
	 * 
	 * @return identifier of the agent
	 */
	public AID getWinnerAgent()
	{
		return this.winnerAgent;
	}

	/* (non-Javadoc)
	 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
	 */
	@Override
	public int compare(HTMLTagA o1, HTMLTagA o2)
	{
		return o1.getRank().compareTo(o2.getRank());
	}
	
}
