package pw.stadnik.searcher.algorithms;

import jade.core.AID;

import java.util.*;

import pw.stadnik.common.HTMLTagA;

/**
 * Class represents game theory method algorithm.
 * It contains all necessary methods to 
 * perform the Game Theory method data processing.
 * 
 * @author Jakub Stadnik
 *
 */
public class GameTheoryMethod
{
	/**
	 * Map of keep payoffs
	 */
	private Map<AID, Float> keepPayoff = new HashMap<AID, Float>();
	/**
	 * Map of change payoffs
	 */
	private Map<AID, Float> changePayoff = new HashMap<AID, Float>();
	/**
	 * Map of links ranking
	 */
	private Map<AID, List<HTMLTagA>> linksRanking = new HashMap<AID, List<HTMLTagA>>();
	/**
	 * First agent chosen in the round
	 * of negotiation
	 */
	private AID firstAgent;
	/**
	 * Second agent chosen in the round
	 * of negotiation
	 */
	private AID secondAgent;
	/**
	 * Winner of the round of negotiation
	 */
	private AID winner;
	/**
	 * Loser of the round of negotiation
	 */
	private AID loser;
	//private Set<AID> agents;
	/**
	 * Highest ranked link of the first agent
	 */
	private HTMLTagA firstAgentLink;
	/**
	 * Highest ranked link of the second agent
	 */
	private HTMLTagA secondAgentLink;
	/**
	 * Link that won the whole negotiation
	 */
	private HTMLTagA winnerLink;
	/**
	 * Copy of full links ranking
	 */
	private Map<AID, List<HTMLTagA>> finalLinksRanking = null;
	/**
	 * Initial search results
	 */
	private Map<AID, List<HTMLTagA>> initialResults = null;
	/**
	 * Returns the link that won the whole negotiation
	 * 
	 * @return link that won the negotiation
	 */
	public HTMLTagA getWinnerLink()
	{
		return this.winnerLink;
	}
	
	/**
	 * Calculates and stores keep payoff for <code>agent</code>
	 * 
	 * @param agent agent for which the keep payoff should be calculated
	 * @param agentAnswerConfidence confidence of agent answer
	 * @param contesterAnswerConfidence confidence of contester about agent answer
	 */
	private void setKeepPayoff(AID agent, Float agentAnswerConfidence, Float contesterAnswerConfidence)
	{
		keepPayoff.put(agent, agentAnswerConfidence - contesterAnswerConfidence);
	}
	/**
	 * Calculates and stores change payoff for <code>agent</code>
	 * 
	 * @param agent agent for which the change payoff should be calculated
	 * @param agentAnswerConfidence confidence of agent answer
	 * @param contesterAnswerConfidence confidence of contester about agent answer
	 */
	private void setChangePayoff(AID agent, Float agentAnswerConfidence, Float contesterAnswerConfidence)
	{
		changePayoff.put(agent, (agentAnswerConfidence + contesterAnswerConfidence) / 2.0f);
	}
	
	/**
	 * Finds the number of agent still in the negotiation
	 * 
	 * @return number of agents in negotiation
	 */
	public Integer getAgentCount()
	{
		return linksRanking.keySet().size();
	}
	
	/**
	 * Method invokes helper method from
	 * AlgorithmHelper class which in turn
	 * calculates weights.
	 * 
	 * @return map of weights calculated by helper
	 */
	public Map<AID, Float> calculateWeights()
	{
		return AlgorithmHelper.calculateWeights(linksRanking, initialResults);
	}
	
	/**
	 * Method updates links ranking to be used after removal
	 * of link which is already in the answer list.
	 * 
	 * @param ranking ranking after removal of link
	 */
	@SuppressWarnings("unchecked")
	public void updateRanking(Map<AID, List<HTMLTagA>> ranking)
	{
		this.linksRanking = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)ranking).clone();
	}
	
	/**
	 * Method determines which action agent will
	 * choose in the round.
	 * 
	 * @param agent agent for which the action should be determined
	 * @return true if the action is keep action, false otherwise
	 */
	private boolean determineAgentAction(AID agent)
	{
		if(this.keepPayoff.get(agent) < this.changePayoff.get(agent))
			return false;
		return true;
	}
	
	/**
	 * Method determines the winner of the
	 * single round of the negotiation.
	 * An agent which chosen the keep action is the winner,
	 * the one that chosen change is loser.
	 */
	private void determineRoundWinner()
	{
		boolean firstAgentAction;
		boolean secondAgentAction;
		if(secondAgent == null)
		{
			winner = firstAgent;
			winnerLink = firstAgentLink;
		}
		else
		{
			firstAgentAction = determineAgentAction(firstAgent);
			secondAgentAction = determineAgentAction(secondAgent);
			if(firstAgentAction && !secondAgentAction)
			{
				winner = firstAgent;
				loser = secondAgent;
				winnerLink = firstAgentLink;
				//this.linksRanking.get(winner).put(winnerLink, this.keepPayoff.get(winner));
			}
			else if(!firstAgentAction && secondAgentAction)
			{
				winner = secondAgent;
				loser = firstAgent;
				winnerLink = secondAgentLink;
			}
			else if((firstAgentAction == secondAgentAction))
			{
				Float firstInitialRanking = finalLinksRanking.get(firstAgent).get(
						finalLinksRanking.get(firstAgent).indexOf(firstAgentLink)).getRank();
				Float secondInitialRanking = finalLinksRanking.get(secondAgent).get(
						finalLinksRanking.get(secondAgent).indexOf(secondAgentLink)).getRank();
				if(firstInitialRanking >= secondInitialRanking)
				{
					winner = firstAgent;
					loser = secondAgent;
					winnerLink = firstAgentLink;
				}
				else
				{
					winner = secondAgent;
					loser = firstAgent;
					winnerLink = secondAgentLink;
				}
				/*
				firstAgentLink.setRank((equalAction == true) ? this.keepPayoff.get(firstAgent) : this.changePayoff.get(firstAgent));
				//this.linksRanking.get(firstAgent).add(firstAgentLink);
				secondAgentLink.setRank((equalAction == true) ? this.keepPayoff.get(secondAgent) : this.changePayoff.get(secondAgent));
				//this.linksRanking.get(secondAgent).add(secondAgentLink);
				sameActionCounter++;
				if(sameActionCounter == 2)
				{
					sameActionCounter = 0;
					if(equalAction == true)
					{
						Float maxKeepPayoff = Float.MIN_VALUE;
						for(AID sa : this.keepPayoff.keySet())
						{
							if(maxKeepPayoff < this.keepPayoff.get(sa))
							{
								maxKeepPayoff = this.keepPayoff.get(sa);
								winner = sa;
								winnerLink = findHighestLink(this.linksRanking.get(winner));
							}
						}
						for(AID sa : this.keepPayoff.keySet())
						{
							if(sa == winner)
								continue;
							loser = sa;
						}
					}
					else
					{
						Float maxChangePayoff = Float.MIN_VALUE;
						for(AID sa : this.changePayoff.keySet())
						{
							if(maxChangePayoff < this.changePayoff.get(sa))
							{
								maxChangePayoff = this.changePayoff.get(sa);
								loser = sa;
							}
						}
						for(AID sa : this.changePayoff.keySet())
						{
							if(sa == loser)
								continue;
							winner = sa;
							winnerLink = findHighestLink(this.linksRanking.get(winner));
						}
					}
				}
				else
					return;*/
			}
		}
		winnerLink.setRank(this.keepPayoff.get(winner));
		//this.linksRanking.get(winner).put(winnerLink, );
		linksRanking.remove(loser);
	}
	
	/**
	 * Method invokes the helper method from
	 * AlgorithmHelper class which in turn finds
	 * the links ranking.
	 * 
	 * @param results initial search results
	 * @param currentWeights currently assigned weights
	 * @return true if links were properly ranked (no adjacent result sets), false otherwise
	 */
	@SuppressWarnings("unchecked")
	public boolean filterAndRankResults(Map<AID, List<HTMLTagA>> results, Map<AID, Float> currentWeights)
	{	
		this.initialResults = (Map<AID, List<HTMLTagA>>)((HashMap<AID, List<HTMLTagA>>)results).clone();
		this.finalLinksRanking = new HashMap<AID, List<HTMLTagA>>();
		this.linksRanking = AlgorithmHelper.filterAndRankResults(results, this.finalLinksRanking, currentWeights);
		if(linksRanking == null)
			return false;
		return true;
	}
	
	/**
	 * Returns full links ranking
	 * 
	 * @return links ranking for each agent
	 */
	public Map<AID, List<HTMLTagA>> getRanking()
	{
		return this.finalLinksRanking;
	}
	
	/**
	 * Method invokes next round of negotiation
	 */
	public void nextRound()
	{
		this.getFirstAgent();
		this.getSecondAgent();
		this.prepareMatrices();
		this.determineRoundWinner();
	}
	
	/**
	 * Method finds the agent which link
	 * is the highest ranked link of all
	 * links from all result sets.
	 */
	private void getFirstAgent()
	{
		//AID first = null;
		Float highestRank = Float.MIN_VALUE;
		for(AID agent : linksRanking.keySet())
		{
			List<HTMLTagA> links = linksRanking.get(agent);
			HTMLTagA tagA = findHighestLink(links);
			Float f = tagA.getRank();
			if(f != null)
			{
				if(highestRank < f)
				{
					highestRank = f;
					firstAgent = agent;
					firstAgentLink = tagA;
				}
			}
		}
		//return first;
	}
	/**
	 * Method finds the agent which link
	 * is the second highest ranked link of all
	 * links from all result sets.
	 */
	private void getSecondAgent()
	{
		// second = null;
		Float highestRank = Float.MIN_VALUE;
		for(AID agent : linksRanking.keySet())
		{
			if(agent.equals(firstAgent))
				continue;
			List<HTMLTagA> links = linksRanking.get(agent);
			HTMLTagA tagA = findHighestLink(links);
			Float f = tagA.getRank();
			if(f != null)
			{
				if(highestRank < f)
				{
					highestRank = f;
					//second = agent;
					secondAgent = agent;
					secondAgentLink = tagA;
				}
			}
		}
		//return second;
	}
	
	//private Float findContesterAnswerConfidence(HTMLTagA contesterAnswer, HashMap<Float, HTMLTagA> )
	
	/**
	 * Method invokes keep and payoff matrices calculation
	 */
	private void prepareMatrices()
	{
		if(secondAgent == null)
			return;
		else
		{
			keepPayoff.clear();
			changePayoff.clear();
			Float firstAgentConfidence = linksRanking.get(firstAgent).get(linksRanking.get(firstAgent).indexOf(firstAgentLink)).getRank();
			Float secondAgentConfidence = linksRanking.get(secondAgent).get(linksRanking.get(secondAgent).indexOf(secondAgentLink)).getRank();
			Float firstAgentContesterAnswerConfidence = linksRanking.get(firstAgent).get(linksRanking.get(firstAgent).indexOf(secondAgentLink)).getRank();
			Float secondAgentContesterAnswerConfidence = linksRanking.get(secondAgent).get(linksRanking.get(secondAgent).indexOf(firstAgentLink)).getRank();
			this.setKeepPayoff(firstAgent, firstAgentConfidence, firstAgentContesterAnswerConfidence);
			this.setKeepPayoff(secondAgent, secondAgentConfidence, secondAgentContesterAnswerConfidence);
			this.setChangePayoff(firstAgent, firstAgentConfidence, firstAgentContesterAnswerConfidence);
			this.setChangePayoff(secondAgent, secondAgentConfidence, secondAgentContesterAnswerConfidence);
		}
	}
	
	/**
	 * Method finds the highest ranked link 
	 * in the ranked result set
	 * 
	 * @param links ranked result set
	 * @return highest ranked link
	 */
	private HTMLTagA findHighestLink(List<HTMLTagA> links)
	{
		Float highestRank = Float.MIN_VALUE;
		HTMLTagA highestLink = null;
		for(HTMLTagA tagA : links)
		{
			if(tagA.getRank() > highestRank)
			{
				highestLink = tagA;
				highestRank = tagA.getRank();
			}
		}
		return highestLink;
	}
}
