package pw.stadnik.searcher.searchengine;

import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.List;

import pw.stadnik.common.HTMLTagA;

/**
 * The class stores data and configuration 
 * of a search engine
 * 
 * @author Jakub Stadnik
 */
public class SearchEngine implements Serializable
{
	/**
	 * Serial version UID. It is used when 
	 * serializing a particular object of this class.
	 */
	private static final long serialVersionUID = -2912745400418952076L;

	/**
	 * Name of the search engine.
	 */
	protected String name;

	/**
	 * Database identifier of the search engine
	 */
	protected Integer id;
	
	/**
	 * Rank of the search engine
	 */
	protected Integer rank;
	
	/**
	 * List of links that should be ignored
	 * for this search engine
	 */
	protected List<HTMLTagA> ignoreLinks;
	
	/**
	 * Parameter representing URL of the search engine
	 */
	protected String queryParameter;
	
	/**
	 * HTTP request parameter representing the results
	 * per page count
	 */
	protected String resultCountParameter;
	
	/**
	 * Results per page stored in the cookie
	 */
	private String cookieBasedParam;
	
	/**
	 * Cookie to set when querying search engine
	 */
	private String cookie = null;
	
	/**
	 * Method constructs query using given search engine
	 * 
	 * @param phrases phrases to include in the query
	 * @param resultCount result count HTTP request parameter
	 * @return URL ready to query the search engine
	 * @throws MalformedURLException if the URL was not correctly constructed
	 */
	public URL constructQuery(List<String> phrases, Integer resultCount) throws MalformedURLException
	{
		StringBuilder query = new StringBuilder();
		
		for(int i = 0; i < phrases.size(); i++)
		{
			try
			{
				query.append(URLEncoder.encode(phrases.get(i),"UTF-8") + "+");
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
		}
		query = new StringBuilder(query.substring(0, query.length() - 1));
		
		//query = new StringBuilder(query.toString().replace("%2B", "+"));
		
		query.insert(0, this.queryParameter);
		//query = new StringBuilder(subString);
		if(!cookieBasedParam.equals("0"))
			constructCookie(resultCount);
		else
			query.append("&" + this.resultCountParameter + "=" + resultCount);
		URL url = new URL(query.toString());
		return url;
	}
	
	/**
	 * Method retrieves the list of links
	 * that should be ignored when
	 * parsing the result page of <code>this</code>
	 * search engine.
	 */
	public void getIgnoreData()
	{
		SearchEngineDataAccess dataAccess = new SearchEngineDataAccess();
		dataAccess.openConnection();
		this.ignoreLinks = dataAccess.getIgnoreData(this.id);
		dataAccess.closeConnection();
	}
	
	/**
	 * Returns the database identifier of the search 
	 * engine
	 * 
	 * @return database identifier
	 */
	public Integer getId()
	{
		return id;
	}

	/**
	 * Returns the list of links that should be ignored
	 * 
	 * @return list of links
	 */
	public List<HTMLTagA> getIgnoreLinks()
	{
		return ignoreLinks;
	}

	/**
	 * Returns the name of the search engine
	 * 
	 * @return name of the search engine
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * Returns the parameter used to construct the query
	 * 
	 * @return query parameter
	 */
	public String getQueryParameter()
	{
		return queryParameter;
	}

	/**
	 * Returns the rank of the search engine
	 * 
	 * @return rank of the search engine
	 */
	public Integer getRank()
	{
		return rank;
	}

	/**
	 * Returns the HTTP request parameter used to set
	 * the results per page count
	 * 
	 * @return HTTP request parameter
	 */
	public String getResultCountParameter()
	{
		return resultCountParameter;
	}

	/**
	 * Sets the database identifier of the search 
	 * engine
	 * 
	 * @param id new database identifier
	 */
	public void setId(Integer id)
	{
		this.id = id;
	}

	/**
	 * Sets the list of links that should be ignored
	 * 
	 * @param ignoreLinks new list of links 
	 */
	public void setIgnoreLinks(List<HTMLTagA> ignoreLinks)
	{
		this.ignoreLinks = ignoreLinks;
	}

	/**
	 * Sets the name of the search engine
	 * 
	 * @param name new name of the search engine
	 */
	public void setName(String name)
	{
		this.name = name;
	}

	/**
	 * Sets the parameter used to construct the query
	 * 
	 * @param queryParameter new query parameter
	 */
	public void setQueryParameter(String queryParameter)
	{
		this.queryParameter = queryParameter;
	}

	/**
	 * Sets the rank of the search engine
	 * 
	 * @param rank new rank of the search engine
	 */
	public void setRank(Integer rank)
	{
		this.rank = rank;
	}

	/**
	 * Sets the HTTP request parameter used to set
	 * the results per page count
	 * 
	 * @param resultCountParameter new parameter
	 */
	public void setResultCountParameter(String resultCountParameter)
	{
		this.resultCountParameter = resultCountParameter;
	}

	/**
	 * Returns parameter used to store
	 * the count of results per page
	 * in cookie
	 * 
	 * @return the cookie parameter
	 */
	public String getCookieBasedParam()
	{
		return cookieBasedParam;
	}

	/**
	 * Sets parameter used to store
	 * the count of results per page
	 * in cookie
	 * 
	 * 
	 * @param cookieBasedParam new cookie parameter
	 */
	public void setCookieBasedParam(String cookieBasedParam)
	{
		this.cookieBasedParam = cookieBasedParam;
	}

	/**
	 * Returns the cookie
	 * 
	 * @return cookie
	 */
	public String getCookie()
	{
		return cookie;
	}

	/**
	 * Sets the coookie
	 * 
	 * @param cookie new cookie
	 */
	public void setCookie(String cookie)
	{
		this.cookie = cookie;
	}
	
	/**
	 * Method inserts into cookie that is
	 * to be used during querying the number
	 * of results that should be displayed on one page.
	 * 
	 * @param resultCount number of results to display
	 */
	private void constructCookie(Integer resultCount)
	{
		this.cookie = this.cookieBasedParam;
		int countParamIndex = this.cookie.indexOf(resultCountParameter);
		String replacementString = this.cookie.substring(countParamIndex);
		String stringToReplace = replacementString = replacementString.substring(0, replacementString.indexOf("&"));
		int equalsSignIndex = replacementString.indexOf("=");
		if(this.name.equals("Ask.com"))
		{
			if(resultCount <= 10)
			{
				replacementString = replacementString.substring(0, equalsSignIndex) + "=0";
			}
			else if(resultCount > 10 && resultCount <= 20)
			{
				replacementString = replacementString.substring(0, equalsSignIndex) + "=1";
			}
			else if(resultCount > 20 && resultCount <= 30)
			{
				replacementString = replacementString.substring(0, equalsSignIndex) + "=2";
			}
			else if(resultCount > 30 && resultCount <= 50)
			{
				replacementString = replacementString.substring(0, equalsSignIndex) + "=3";
			}
			else if(resultCount > 50 && resultCount <= 100)
			{
				replacementString = replacementString.substring(0, equalsSignIndex) + "=4";
			}
		}
		else
		replacementString = replacementString.substring(0, equalsSignIndex) + "=" + resultCount;
		this.cookie = this.cookie.replace(stringToReplace, replacementString);
	}
	
}
